"""
    The cpb module gives you access to a binary representatin of 
    digital data objects hosted at the ICOS CarbonPortal.
    Any data column (x,y axis) you can "preview" on the data portal
    ( https://data.icos-cp.eu/portal/ ) is accessible from this module.
    You need to know the digital object ID (URL, persistent identification)
    or have a sqparql query providing this information.    
"""
